# coding=utf-8
from __future__ import absolute_import

###
# Print It Cloud integration plugin

import os
import io
import base64
from time import sleep

import socketio
import requests
from octoprint_print_it_cloud.pic.EventBacklogProcessor import EventBacklogProcessor
from octoprint_print_it_cloud.pic.LoggingMixin import LoggingMixin

from octoprint.util import RepeatedTimer
from datetime import datetime
import octoprint.plugin
from flask import jsonify
from octoprint.events import Events
from octoprint.filemanager.destinations import FileDestinations
import json

PUSH_STATUS_SLOW = 60
PUSH_STATUS_NORMAL = 10
PUSH_STATUS_FAST = 1


class PrintItCloudPlugin(octoprint.plugin.StartupPlugin,
						 octoprint.plugin.ShutdownPlugin,
						 octoprint.plugin.SimpleApiPlugin,
						 octoprint.plugin.SettingsPlugin,
						 octoprint.plugin.AssetPlugin,
						 octoprint.plugin.TemplatePlugin,
						 octoprint.plugin.EventHandlerPlugin,
						 LoggingMixin):
	_backlog_timer: RepeatedTimer
	_status_timer: RepeatedTimer
	_backlog_processor: EventBacklogProcessor
	_connectivity_check_timer: RepeatedTimer

	def __init__(self):
		super().__init__()
		self._sio = socketio.Client(reconnection=True, reconnection_delay=5, reconnection_delay_max=20)

		self._snapshot_url = None
		self._snapshot_capture_errors = 0
		self._snapshot_capture_skip = 0

		self._backlog_timer = None
		self._backlog_processor = None

		self._syncing = False
		self._connecting = False
		self._manually_disconnected = False
		self._connectivity_check_timer = None
		self._status_timer = None
		self._status_push_interval = PUSH_STATUS_SLOW
		self._last_client_alive = datetime.now()

	# StartupPlugin
	def on_after_startup(self):
		self._log_info("Print It Cloud plugin initialized")
		self._snapshot_url = self._settings.global_get(['webcam', 'snapshot'])
		if not self._file_manager.folder_exists(FileDestinations.LOCAL, 'printit.cloud'):
			self._file_manager.add_folder(FileDestinations.LOCAL, 'printit.cloud')

		try:
			self._connect_io_events()
			self._setup_backlog_timer()
			self._setup_connectivity_check_timer()
		except Exception as e:
			self._log_error("Failed initializing the cloud plugin".format(e.__str__()))

	# END on_after_startup

	# The getter is needed in order for the timer to use actual interval defined by self._status_push_interval
	def _get_status_push_interval(self):
		return self._status_push_interval

	def _remove_status_push_timer(self):
		if self._status_timer:
			try:
				self._log_info("Removing the status timer")
				self._status_timer.cancel()
				self._status_timer = None
			except Exception as e:
				self._log_error("Failed removing the status timer: {}".format(e.__str__()))

	def _setup_status_push_timer(self, interval):
		self._log_debug("Setting up the status timer with interval {}".format(interval))
		if self._status_timer and (interval == self._status_push_interval):
			return

		if self._status_timer:
			self._status_timer.cancel()

		self._status_push_interval = interval
		self._status_timer = RepeatedTimer(self._get_status_push_interval, self._push_status_to_cloud)
		self._status_timer.start()
		self._log_debug("The status timer started")

	def _setup_connectivity_check_timer(self):
		if self._connectivity_check_timer:
			return
		self._log_debug("Setting up connectivity check timer")
		self._connectivity_check_timer = RepeatedTimer(60, self._on_connectivity_check_timer)
		self._connectivity_check_timer.start()
		self._log_debug("Connectivity check timer started")
		self._on_connectivity_check_timer()

	def _setup_backlog_timer(self):
		if self._backlog_timer and self._backlog_timer.is_alive():
			return
		self._log_debug("Setting up backlog sync timer")
		self._backlog_timer = RepeatedTimer(5, self._sync_backlog)
		self._backlog_timer.start()
		self._log_debug("Backlog sync timer started")

	def _sync_backlog(self):
		count = self._backlog_processor.count()

		self._syncing = count.__bool__()
		self.fire_state_to_ui()

		if not count:
			return

		id, event, data, timestamp = self._backlog_processor.next()
		self._safe_emit('printer_backlog_event', dict(
			name=event,
			data=data,
			timestamp=timestamp,
			data_id=id
		), '/printers')
		self._log_info("HUB_BACKLOG_SYNC_EVENT sent \"{}\".".format(id))
		pass

	def _on_connectivity_check_timer(self):
		if self._settings.get(["autoConnect"]) and not self._manually_disconnected:
			self._async_connect_to_cloud()

	# connect_to_cloud does nothing if already connected
	def _async_connect_to_cloud(self):
		self._sio.start_background_task(self._connect_to_cloud)

	def _connect_to_cloud(self):
		if self._connecting:
			self._log_debug('_connect_to_cloud: printer is connecting - skipping')
			return

		api_key = self._settings.get(["apiKey"])

		if (not self._sio.connected) and (not self._manually_disconnected) and api_key:
			try:
				self._log_debug('_connect_to_cloud: set connecting True')
				self._connecting = True

				hub_base_url = self._settings.get(["hubBaseUrl"])
				io_url = "{}?apiKey={}&v={}".format(hub_base_url, api_key, self._plugin_version)
				self._log_debug('_connect_to_cloud: api_key is set - connecting')
				self._log_info("Connecting to the cloud: " + io_url)
				self._sio.connect(
					io_url,
					headers={},
					transports=None,
					# transports=['websocket'],
					namespaces=['/printers'],
					socketio_path='/api/v1.0/public/hub/socket.io'
				)
			except Exception as e:
				self._connecting = False
				self._log_error("Failed connecting to the cloud: {}".format(e.__str__()))
		pass

	def _connect_io_events(self):
		@self._sio.event(namespace='/printers')
		def connect():
			self._connecting = False
			self._say_welcome()
			# noinspection PyUnresolvedReferences
			self._event_bus.fire(Events.PLUGIN_PRINT_IT_CLOUD_MESSAGE, dict(type="connect", payload=True))
			try:
				self._last_client_alive = datetime.now()
				# Let's go fast initially to show up the printer connected asap
				self._setup_status_push_timer(PUSH_STATUS_FAST)
				# Report the status right away to do not wait for the timer
				self._sio.start_background_task(self._push_status_to_cloud)
			except Exception as e:
				self._log_error("Failed starting to push status to the cloud: {}".format(e.__str__()))

		@self._sio.event(namespace='/printers')
		def connect_error():
			self._connecting = False
			# noinspection PyUnresolvedReferences
			self._event_bus.fire(Events.PLUGIN_PRINT_IT_CLOUD_MESSAGE, dict(type="connect_error", payload=False))
			self._remove_status_push_timer()
			self._log_info("Connection to the cloud failed! (connect_error)")

		@self._sio.event(namespace='/printers')
		def disconnect():
			self._connecting = False
			# noinspection PyUnresolvedReferences
			self._event_bus.fire(Events.PLUGIN_PRINT_IT_CLOUD_MESSAGE, dict(type="disconnect", payload=False))
			self._remove_status_push_timer()
			self._log_info("Disconnected form the cloud!")

		@self._sio.event(namespace='/printers')
		def hub_disconnect_event():
			self._sio.disconnect()

		@self._sio.event(namespace='/printers')
		def hub_report_status_event():
			self._last_client_alive = datetime.now()
			self._sio.start_background_task(self._push_status_to_cloud)

		@self._sio.event(namespace='/printers')
		def hub_command_event(event):
			self._log_info('hub_command_event event received')
			self._last_client_alive = datetime.now()
			try:
				if not ('module' in event):
					return dict(
						result=False,
						msg='Bad event. Missing module name'
					)

				if not ('cmd' in event):
					return dict(
						result=False,
						msg='Bad event. Missing cmd name'
					)

				if event['module'] == 'cloud':
					if event['cmd'] == 'complete_task':
						return self._handle_complete_task_event(event)

					if event['cmd'] == 'print':
						return self._handle_print_task_event(event)

					return dict(
						result=False,
						msg='Unknown command given for module cloud: ' + event['cmd']
					)

				# Octoprint internal modules
				if not (hasattr(self, event['module'])):
					return dict(
						result=False,
						msg='Unknown module: ' + event['module']
					)

				module = getattr(self, event['module'])
				if not (hasattr(module, event['cmd']) and callable(getattr(module, event['cmd']))):
					return dict(
						result=False,
						msg='Unknown command: ' + event['cmd']
					)

				if "args" in event:
					_func = getattr(module, event['cmd'])
					result = _func(**event['args'])
				else:
					result = getattr(module, event['cmd'])()

				self._settings.save(True)

				if result:
					return dict(
						result=True,
						data=result
					)
				else:
					return dict(result=True)
			except Exception as e:
				self._log_error("Failed to execute command")
				self._log_error(event)
				self._log_error(e)
				return dict(
					result=False,
					error=e.__str__()
				)

		@self._sio.event(namespace='/printers')
		def hub_pong_event(msg):
			self._log_info("PONG received \"{}\". Updating the alive timer: ".format(msg))
			self._last_client_alive = datetime.now()

		@self._sio.event(namespace='/printers')
		def hub_backlog_read_event(msg):
			data_id = int(msg)
			if msg:
				self._log_info("HUB_BACKLOG_READ_EVENT received \"{}\".".format(data_id))
				self._backlog_processor.rm(data_id)
				self._sio.start_background_task(self._sync_backlog)

		@self._sio.event(namespace='/printers')
		def hub_set_speed_event(speed="SLOW"):
			# Reset last alive time
			self._last_client_alive = datetime.now()
			if speed == 'SLOW':
				self._setup_status_push_timer(PUSH_STATUS_SLOW)

			if speed == 'NORMAL':
				self._setup_status_push_timer(PUSH_STATUS_NORMAL)

			if speed == 'FAST':
				self._setup_status_push_timer(PUSH_STATUS_FAST)

			# Report the status right away to do not wait for the timer
			self._sio.start_background_task(self._push_status_to_cloud)

	def on_shutdown(self):
		self._remove_status_push_timer()
		# close socket connection
		if self._sio.connected:
			self._sio.disconnect()
		self._say_bye()

	def _slow_down_inactive(self):
		#  Set to low speed if no clients are alive
		if (datetime.now() - self._last_client_alive).total_seconds() > 30:
			if self._status_push_interval != PUSH_STATUS_SLOW:
				self._log_debug("RESET SPEED TO SLOW")
				self._setup_status_push_timer(PUSH_STATUS_SLOW)

	def _safe_emit(self, event, data=None, namespace=None, callback=None):
		try:
			if self._sio.connected:
				if '/printers' in self._sio.connection_namespaces:
					self._sio.emit(event, data, namespace, callback)
		except Exception as e:
			self._log_debug('Failed to emit {}: {}'.format(event, e.__str__()))

	def _push_status_to_cloud(self):
		self._safe_emit('printer_status_report_event', dict(
			img=self._capture_webcam_snapshot(),
			version=self._plugin_version,
			temperatures=self._printer.get_current_temperatures(),
			**self._printer.get_current_data()
		), namespace='/printers')
		self._slow_down_inactive()

	def _handle_complete_task_event(self, event):
		if not ('args' in event):
			return dict(
				result=False,
				msg='Missing args'
			)
		payload = event['args']
		if not payload['taskId']:
			return dict(
				result=False,
				msg='Bad args. Missing taskId'
			)

		self._printer.unselect_file()
		task_file_path = 'printit.cloud/{}.gcode'.format(payload['taskId'])
		if self._file_manager.file_exists(FileDestinations.LOCAL, task_file_path):
			self._file_manager.remove_file(FileDestinations.LOCAL, task_file_path)

		return dict(
			result=True,
			msg='Task completed'
		)

	def _handle_print_task_event(self, event):
		if not ('args' in event):
			return dict(
				result=False,
				msg='Missing args'
			)
		args = event['args']
		if not args['gcodeId']:
			return dict(
				result=False,
				msg='Bad args. Missing gcodeId'
			)
		if not args['_id']:
			return dict(
				result=False,
				msg='Bad args. Missing task _id'
			)
		if not self._printer.is_operational():
			return dict(
				result=False,
				msg='Printer is not operational'
			)
		if self._printer.is_closed_or_error():
			return dict(
				result=False,
				msg='Printer is in error state'
			)
		if self._printer.is_printing() or self._printer.is_cancelling() \
				or self._printer.is_pausing() or self._printer.is_paused() \
				or self._printer.is_error() or not self._printer.is_ready():
			return dict(
				result=False,
				msg='Printer is busy. Can not accept jobs'
			)

		import requests
		url = 'https://storage.printit.cloud/gcodes/{}.gcode'.format(args['gcodeId'])
		response = requests.get(url, stream=True, timeout=(10.0, 60.0))

		if response.status_code == 200:
			if not self._file_manager.folder_exists(FileDestinations.LOCAL, 'printit.cloud'):
				self._file_manager.add_folder(FileDestinations.LOCAL, 'printit.cloud')

			# Cleanup existing of prev fails (if any)
			task_file_path = 'printit.cloud/{}.gcode'.format(args['_id'])
			if self._file_manager.file_exists(FileDestinations.LOCAL, task_file_path):
				self._file_manager.remove_file(FileDestinations.LOCAL, task_file_path)

			path, name = self._file_manager.sanitize(FileDestinations.LOCAL, task_file_path)
			folder_path = os.path.join(path, name)
			with open(folder_path, 'wb') as file:
				for chunk in response.iter_content(100000):  # download 100kb at a time
					file.write(chunk)

			self._printer.select_file(task_file_path, False)

			if not self._printer.is_current_file(task_file_path, False):
				return dict(
					result=False,
					msg='Error selecting file to print'
				)

			self._printer.start_print()

			return dict(
				result=self._printer.is_printing(),
				msg='Print start cmd executed'
			)
		else:
			return dict(
				result=False,
				msg='G-CODE downloading error'
			)
		pass

	def _capture_webcam_snapshot(self):
		if self._snapshot_capture_errors >= 3:
			self._snapshot_capture_skip += 1
			# Lets try capturing again after some number of skipping
			if self._snapshot_capture_skip >= 50:
				self._snapshot_capture_skip = 0
				self._snapshot_capture_errors = 0
			return None

		if self._snapshot_url is not None and self._snapshot_url.strip() != "":
			try:
				self._log_debug("Capture snapshot from {}".format(self._snapshot_url))
				req = requests.get(
					self._snapshot_url,
					stream=True,
					timeout=300,
					verify=False,
				)
				req.raise_for_status()
				self._snapshot_capture_errors = 0
				return "data:image/png;base64," + base64.b64encode(req.content).decode("utf8")
			except Exception as e:
				self._snapshot_capture_errors += 1
				self._log_error("Could not capture image from {}".format(self._snapshot_url))
				if self._snapshot_capture_errors >= 3:
					self._log_debug('Start skipping snapshots capturing due to too many errors')
				return None
		return None

	# SimpleApiPlugin
	def get_api_commands(self):
		return dict(
			connect=[],
			disconnect=[]
		)

	def on_api_command(self, command, data):
		self._log_info(command)
		api_key = self._settings.get(["apiKey"])
		if not api_key:
			return jsonify(dict(
				connected=self._sio.connected,
				configured=bool(api_key)
			))

		if command == "connect":
			self._manually_disconnected = False
			if self._sio.connected:
				self.fire_state_to_ui()
				return jsonify(dict(
					connected=self._sio.connected,
					configured=bool(api_key)
				))
			try:
				self._async_connect_to_cloud()
				self.fire_state_to_ui()
				return jsonify(dict(
					connected=self._sio.connected,
					configured=bool(api_key)
				))
			except Exception as e:
				self._log_error(e.__str__())
				response = jsonify({
					'status': 500,
					'sub_code': 901,
					'message': "Не удалось подключиться к облаку"
				})
				response.status_code = 500
				self.fire_state_to_ui()
				return response

		if command == "disconnect":
			self._manually_disconnected = True

			if not self._sio.connected:
				self.fire_state_to_ui()
				return jsonify(dict(
					connected=self._sio.connected,
					configured=bool(api_key)
				))
			try:
				self._sio.disconnect()
				self.fire_state_to_ui()
				return jsonify(dict(
					connected=self._sio.connected,
					configured=bool(api_key)
				))
			except Exception as e:
				self._log_error("Failed disconnecting from the cloud")
				self._log_error(e.__str__())
				response = jsonify({
					'status': 500,
					'sub_code': 902,
					'message': "Оибка запроса"
				})
				response.status_code = 500
				self.fire_state_to_ui()
				return response

	def fire_state_to_ui(self):
		self._event_bus.fire(
			Events.PLUGIN_PRINT_IT_CLOUD_MESSAGE,
			dict(type="state_change", payload=dict(
				syncing=self._syncing,
				connected=self._sio.connected,
				configured=bool(self._settings.get(["apiKey"]))
			))
		)

	def on_api_get(self, request):
		self.fire_state_to_ui()
		api_key = self._settings.get(["apiKey"])
		return jsonify(dict(
			connected=self._sio.connected,
			configured=bool(api_key)
		))

	##~~ EventHandlerPlugin API

	def on_event(self, event, args):
		if event == 'plugin_print_it_cloud_message':
			return
		if not self._backlog_processor:
			self._backlog_processor = EventBacklogProcessor(self._plugin_version)
		# self._sio.start_background_task(self._backlog_processor.insert, event, json.dumps(args))
		self._backlog_processor.insert(event, json.dumps(args))
		pass

	# def handle_event(self, event, args):
	# 	if self._sio.connected:
	# 		try:
	# 			self._safe_emit('printer_event', dict(
	# 				name=event,
	# 				data=args,
	# 				printerData=self._printer.get_current_data()
	# 			), namespace='/printers')
	# 		except Exception as e:
	# 			self._log_info("ON_EVENT exception {}".format(e.__str__()))

	def on_sockjs_emit(self, socket, user, message, payload):
		# Just ensuring the connection
		# connect_to_cloud does nothing if already connected
		# if self._settings.get(["autoConnect"]) and (not self._manually_disconnected):
		# 	self._async_connect_to_cloud()
		# if self._sio.connected:
		# 	self._safe_emit('sockjs_event', dict(
		# 		type=message,
		# 		data=payload
		# 	), namespace='/printers')
		return True

	##~~ SettingsPlugin mixin

	def get_settings_defaults(self):
		return dict(
			autoConnect=True,
			apiKey=None,
			hubBaseUrl="https://printit.cloud"
		)

	##~~ AssetPlugin mixin

	def get_assets(self):
		# Define your plugin's asset files to automatically include in the
		# core UI here.
		return dict(
			js=[
				"js/bundle.js"
			],
			css=[
				"css/app.css"
			]
		)

	##~~ Softwareupdate hook

	def get_update_information(self):
		hub_base_url = self._settings.get(["hubBaseUrl"])
		json_data = "{}/api/v1.0/public/hub/static/plugin_version.json".format(hub_base_url)
		pip_url = hub_base_url + "/api/v1.0/public/hub/static/PrintIt.Cloud-{target_version}.tar.gz"
		# Define the configuration for your plugin to use with the Software Update
		# Plugin here. See https://docs.octoprint.org/en/master/bundledplugins/softwareupdate.html for details.
		return dict(
			print_it_cloud=dict(
				displayName="PrintIt.Cloud Plugin",
				displayVersion=self._plugin_version,
				type="jsondata",
				jsondata=json_data,
				current=self._plugin_version,
				pip=pip_url
			)
		)


# If you want your plugin to be registered within OctoPrint under a different name than what you defined in setup.py
# ("OctoPrint-PluginSkeleton"), you may define that here. Same goes for the other metadata derived from setup.py that
# can be overwritten via __plugin_xyz__ control properties. See the documentation for that.
__plugin_name__ = "PrintIt Cloud integration plugin"

# Starting with OctoPrint 1.4.0 OctoPrint will also support to run under Python 3 in addition to the deprecated
# Python 2. New plugins should make sure to run under both versions for now. Uncomment one of the following
# compatibility flags according to what Python versions your plugin supports!
# __plugin_pythoncompat__ = ">=2.7,<3" # only python 2
# __plugin_pythoncompat__ = ">=3,<4"  # only python 3

__plugin_pythoncompat__ = ">=3.5,<4"  # python 3.5 and 4


def _register_custom_events(*args, **kwargs):
	return ["message"]


def __plugin_load__():
	global __plugin_implementation__
	__plugin_implementation__ = PrintItCloudPlugin()

	global __plugin_hooks__
	__plugin_hooks__ = {
		"octoprint.plugin.softwareupdate.check_config": __plugin_implementation__.get_update_information,
		"octoprint.server.sockjs.emit": __plugin_implementation__.on_sockjs_emit,
		"octoprint.events.register_custom_events": _register_custom_events
	}
